# Souffle - A Datalog Compiler
# Copyright (c) 2022 The Souffle Developers. All rights reserved
# Licensed under the Universal Permissive License v 1.0 as shown at:
# - https://opensource.org/licenses/UPL
# - <souffle root>/licenses/SOUFFLE-UPL.txt

import os
from common import *

args = os.sys.argv
args.pop(0)

if len(args) == 0:
    raise RuntimeError("Missing TEST_NAME")

test_name = args.pop(0)

if len(args) > 0:
    extra_data = args.pop(0)
else:
    extra_data = None

if len(args) != 0:
    raise RuntimeError("Unexpected argument")

sort_file("{}.out".format(test_name))
compare_sorted_file("{}.out".format(test_name))

compare_file("{}.err".format(test_name))

if extra_data == "python" or extra_data == "java":
    file_name = "{}-{}.out".format(test_name, extra_data)
    sort_file(file_name)
    compare_sorted_file(file_name)





# --- JSON_DATA_TEXT variable is inserted before this line ---

## Example of JSON_DATA_TEXT
if not JSON_DATA_TEXT:
    JSON_DATA_TEXT = """{
      "compiler": "/usr/bin/c++",
      "compiler_id": "GNU",
      "compiler_version": "8.3.0",
      "msvc_version": "",
      "includes": "-I/usr/include",
      "std_flag": "-std=c++17",
      "cxx_flags": " -fopenmp",
      "cxx_link_flags": "",
      "release_cxx_flags": "-O3 ",
      "debug_cxx_flags": "-g",
      "definitions": "-DRAM_DOMAIN_SIZE=64 -DUSE_NCURSES -DUSE_LIBZ -DUSE_SQLITE",
      "compile_options": "",
      "link_options": "-pthread -ldl -lstdc++fs /usr/lib/x86_64-linux-gnu/libsqlite3.so /usr/lib/x86_64-linux-gnu/libz.so /usr/lib/x86_64-linux-gnu/libncurses.so",
      "rpaths": "/usr/lib/x86_64-linux-gnu:/usr/lib/x86_64-linux-gnu",
      "outname_fmt": "-o {}",
      "libdir_fmt": "-L{}",
      "libname_fmt": "-l{}",
      "rpath_fmt": "-Wl,-rpath,{}",
      "path_delimiter": ":",
      "exe_extension": "",
      "source_include_dir": "",
      "jni_includes": ""
    }"""

import argparse
import json
import os
import pathlib
import shutil
import subprocess
import sys
import tempfile

# run command and return status object
def launch_command(cmd, descr, verbose=False):
    if verbose:
        sys.stdout.write(cmd + "\n")
    status = subprocess.run(cmd, capture_output=True, text=True, shell=True)
    if status.returncode != 0:
        sys.stdout.write(status.stdout)
        sys.stderr.write(status.stderr)
        raise RuntimeError("Error: {}. Command: {}".format(descr, cmd))
    return status

# run command and return the standard output as a string
def capture_command_output(cmd, descr, verbose=False):
    status = launch_command(cmd, descr, verbose)
    return status.stdout


conf = json.loads(JSON_DATA_TEXT)
OUTNAME_FMT = conf['outname_fmt']
LIBDIR_FMT = conf['libdir_fmt']
LIBNAME_FMT = conf['libname_fmt']
RPATH_FMT = conf['rpath_fmt']
PATH_DELIMITER = conf['path_delimiter']
RPATHS = conf['rpaths'].split(PATH_DELIMITER)
exeext = conf['exe_extension']
SOURCE_INCLUDE_DIR = conf['source_include_dir']
JNI_INCLUDES = conf['jni_includes'].split(PATH_DELIMITER)

workdir = os.getcwd()
scriptdir = pathlib.Path(os.path.dirname(os.path.abspath(__file__)))

parser = argparse.ArgumentParser(description="Compile a C++ source file generated by Souffle")
parser.add_argument('-l', action='append', default=[], metavar='LIBNAME', dest='lib_names', type=str, help="Basename of a functors library. eg: `-l functors` => libfunctors.dll")
parser.add_argument('-L', action='append', default=[], metavar='LIBDIR', dest='lib_dirs', type=lambda p: pathlib.Path(p).absolute(), help="Search directory for functors libraries")
parser.add_argument('-g', action='store_true', dest='debug', help="Debug build type")
parser.add_argument('-s', metavar='LANG', dest='swiglang', choices=["java", "python"], help="use SWIG interface to generate into LANG language")
parser.add_argument('-v', action='store_true', dest='verbose', help="Verbose output")
parser.add_argument('source', nargs='+', metavar='SOURCE', type=lambda p: pathlib.Path(p).absolute(), help="C++ source files")
parser.add_argument('-o', metavar='BINARY', dest='output', type=lambda p: pathlib.Path(p).absolute(), help="Binary file name")

args = parser.parse_args()

if not args.output:
    raise RuntimeError("Missing output file name in souffle-compile")

for f in args.source:
    if not os.path.isfile(f):
        raise RuntimeError("Cannot open source file: '{}'".format(f))


# Check if the input file has a valid extension
for f in args.source:
    extname = f.suffix
    if extname != ".cpp":
        raise RuntimeError("Source file is not a .cpp file: '{}'".format(f))

# Search for Souffle includes directory
souffle_include_dir = None
if (scriptdir / "include" / "souffle").exists():
    souffle_include_dir = scriptdir / "include" / "souffle"
elif (scriptdir / ".." / "include" / "souffle").exists():
    souffle_include_dir = scriptdir / ".." / "include" / "souffle"
elif SOURCE_INCLUDE_DIR and (pathlib.Path(SOURCE_INCLUDE_DIR) / "souffle").exists():
    souffle_include_dir = (pathlib.Path(SOURCE_INCLUDE_DIR) / "souffle")

if args.swiglang:
    if not (souffle_include_dir and (souffle_include_dir / "swig").exists()):
        raise RuntimeError("Cannot find 'souffle/swig' include directory")

    swig_include_dir = (souffle_include_dir / "swig")
    with tempfile.TemporaryDirectory() as tmpdir:
        shutil.copy(swig_include_dir / "SwigInterface.h", tmpdir)
        shutil.copy(swig_include_dir / "SwigInterface.i", tmpdir)

        os.chdir(tmpdir)
        launch_command("swig -c++ -\"{}\" SwigInterface.i".format(args.swiglang), "SWIG generation", verbose=args.verbose)

        if args.swiglang == "python":
            swig_flags = capture_command_output("python3-config --cflags", "Python config", verbose=args.verbose)
            swig_ldflags = capture_command_output("python3-config --ldflags", "Python config", verbose=args.verbose)
            swig_outname = "_SwigInterface.so"
        elif args.swiglang == "java":
            swig_flags = " ".join(["-I{}".format(dir) for dir in JNI_INCLUDES])
            swig_ldflags = ""
            swig_outname = "libSwigInterface.so"

        # compile swig interface and program
        cmd = []
        cmd.append('"{}"'.format(conf['compiler']))
        cmd.append("-fPIC")
        cmd.append("-c")
        cmd.append("-D__EMBEDDED_SOUFFLE__")
        cmd.append("SwigInterface_wrap.cxx")
        for f in args.source:
            cmd.append(str(f))
        cmd.append(conf['definitions'])
        cmd.append(conf['compile_options'])
        cmd.append(conf['includes'])
        cmd.append(conf['std_flag'])
        cmd.append(conf['cxx_flags'])
        if args.debug:
            cmd.append(conf['debug_cxx_flags'])
        else:
            cmd.append(conf['release_cxx_flags'])
        cmd.append(swig_flags)
        cmd = " ".join(cmd)
        launch_command(cmd, "Compilation of SWIG C++", verbose=args.verbose)

        # link swig interface and program
        cmd = []
        cmd.append('"{}"'.format(conf['compiler']))
        cmd.append("-shared")
        cmd.append("SwigInterface_wrap.o")
        cmd.append(os.path.basename(args.output) + ".o")
        cmd.append("-o")
        cmd.append(swig_outname)
        cmd.append(conf['definitions'])
        cmd.append(conf['compile_options'])
        cmd.append(conf['includes'])
        cmd.append(conf['std_flag'])
        cmd.append(conf['cxx_flags'])
        if args.debug:
            cmd.append(conf['debug_cxx_flags'])
        else:
            cmd.append(conf['release_cxx_flags'])
        cmd.append(conf['link_options'])
        cmd.extend(list(map(lambda rpath: RPATH_FMT.format(rpath), RPATHS)))
        cmd.extend(list(map(lambda libdir: LIBDIR_FMT.format(libdir), args.lib_dirs)))
        cmd.extend(list(map(lambda libname: LIBNAME_FMT.format(libname), args.lib_names)))
        cmd.append(swig_ldflags)
        cmd = " ".join(cmd)
        launch_command(cmd, "Link of SWIG C++", verbose=args.verbose)

        if args.swiglang == "python":
            shutil.copy("_SwigInterface.so", workdir)
            shutil.copy("SwigInterface.py", workdir)
        elif args.swiglang == "java":
            shutil.copy("libSwigInterface.so", workdir)
            for javasrc in pathlib.Path(tmpdir).glob("*.java"):
                shutil.copy(javasrc, workdir)

        # move generated files to same directory as cpp file
        os.sys.exit(0)
else:
    exepath = pathlib.Path("{}{}".format(args.output, exeext))

    cmd = []
    cmd.append('"{}"'.format(conf['compiler']))
    cmd.append(conf['definitions'])
    cmd.append(conf['compile_options'])
    cmd.append(conf['includes'])
    cmd.append(conf['std_flag'])
    cmd.append(conf['cxx_flags'])

    if args.debug:
        cmd.append(conf['debug_cxx_flags'])
    else:
        cmd.append(conf['release_cxx_flags'])

    cmd.append(OUTNAME_FMT.format(exepath))
    for f in args.source:
        cmd.append(str(f))

    cmd.append(conf['link_options'])
    cmd.extend(list(map(lambda rpath: RPATH_FMT.format(rpath), RPATHS)))
    cmd.extend(list(map(lambda libdir: LIBDIR_FMT.format(libdir), args.lib_dirs)))
    cmd.extend(list(map(lambda libname: LIBNAME_FMT.format(libname), args.lib_names)))

    cmd = " ".join(cmd)

    if args.verbose:
        sys.stderr.write(cmd + "\n")

    if exepath.exists():
        exepath.unlink()

    status = subprocess.run(cmd, capture_output=True, text=True, shell=True)
    if status.returncode != 0:
        sys.stdout.write(status.stdout)
        sys.stderr.write(status.stderr)

    os.sys.exit(status.returncode)



#!/usr/bin/env python3

"""Helper for GitHub Actions."""

n_tests, n_chunks, chunk_id = map(int, input().split())
d, m = divmod(n_tests, n_chunks)
print(f"{chunk_id * d + min(m, chunk_id) + 1},{(chunk_id+1) * d + min(m, chunk_id+1)}")



import argparse
import os
import re
import shutil
import subprocess
import sys
import tempfile


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_dir")
    parser.add_argument("--output_dir")
    parser.add_argument("--cxx")
    parser.add_argument("--souffle")
    parser.add_argument("--driver")
    parser.add_argument("--include")
    parser.add_argument("--namespace", action="store_true")
    parser.set_defaults(namespace=False)
    parser.add_argument("dls", nargs="+")

    args = parser.parse_args()
    cxx = args.cxx
    souffle = args.souffle
    driver = args.driver
    include = args.include
    dls = args.dls
    output_dir = args.output_dir

    def cpp_file(dl):
        return re.sub(r"dl$", "cpp", dl)

    def object_file(dl):
        return re.sub(r"dl$", "o", dl)

    os.chdir(args.input_dir)

    if not os.path.isdir(output_dir):
        if os.path.exists(output_dir):
            os.remove(output_dir)
        os.mkdir(output_dir)

    for f in [driver] + dls:
        shutil.copy(f, output_dir)

    os.chdir(output_dir)
    redirects = dict(stdout=sys.stdout, stderr=sys.stderr)
    ns_index = 0
    for dl in dls:
        ns_index += 1
        ns_args = ["-N", "ns" + str(ns_index)] if args.namespace else []
        subprocess.run([souffle] + ns_args + ["-g", cpp_file(dl), dl], **redirects)
    cxx_args = [
        cxx,
        "-I",
        include,
        "--std=c++17",
        "-D__EMBEDDED_SOUFFLE__=1",
    ]
    for dl in dls:
        subprocess.run(
            cxx_args
            + [
                "-c",
                cpp_file(dl),
                "-o",
                object_file(dl),
            ],
            **redirects
        )
    os.execvp(cxx, cxx_args + [driver] + [object_file(dl) for dl in dls])


if __name__ == "__main__":
    main()



# Souffle - A Datalog Compiler
# Copyright (c) 2022 The Souffle Developers. All rights reserved
# Licensed under the Universal Permissive License v 1.0 as shown at:
# - https://opensource.org/licenses/UPL
# - <souffle root>/licenses/SOUFFLE-UPL.txt

import difflib
import re
import os

## sort file `file_name` to file `file_name.sorted`
def sort_file(file_name):
    if not os.path.isfile(file_name):
        raise RuntimeError("Missing file '{}'".format(file_name))

    output_file = "{}.sorted".format(file_name)

    with open(file_name) as f:
        digits = re.compile(r'^\d+$')
        def prepare_row(row):
            return list(map(
                # natural sort, transfor integers to string representation with leading zeroes
                lambda field: (("{:0>20d}".format(int(field))) if digits.match(field) else field),
                row.split("\t")))

        rows = sorted([row.rstrip() for row in f], key=lambda row: prepare_row(row))

    with open(output_file, "w") as f:
        f.write("\n".join(rows))

    return None

## Compare two given files. Exit with error status if files do not match.
def compare_files(expected_file, actual_file):
    if not os.path.isfile(expected_file):
        raise RuntimeError("Missing file '{}'".format(expected_file))
    if not os.path.isfile(actual_file):
        raise RuntimeError("Missing file '{}'".format(actual_file))

    with open(expected_file) as f:
        expected_lines = [line.rstrip() for line in f]

    with open(actual_file) as f:
        actual_lines = [line.rstrip() for line in f]

    if actual_lines != expected_lines:
        os.sys.stdout.writelines(difflib.unified_diff(open(expected_file).readlines(), open(actual_file).readlines(), fromfile=expected_file, tofile=actual_file))
        os.sys.stdout.write("\n")
        os.sys.exit("Found output difference, expected file:'{}', actual file:'{}".format(expected_file, actual_file))

    return True

## Compare `file_name` with `file_name.expected`. Exit with error status if files do not match.
def compare_file(file_name):
    actual_file = file_name
    expected_file = "{}.expected".format(file_name)

    return compare_files(expected_file, actual_file)

## Compare `file_name.sorted` with `file_name.expected.sorted`. Exit with error status if files do not match.
def compare_sorted_file(file_name):
    actual_file = "{}.sorted".format(file_name)
    expected_file = "{}.expected.sorted".format(file_name)

    return compare_files(expected_file, actual_file)




# Souffle - A Datalog Compiler
# Copyright (c) 2022 The Souffle Developers. All rights reserved
# Licensed under the Universal Permissive License v 1.0 as shown at:
# - https://opensource.org/licenses/UPL
# - <souffle root>/licenses/SOUFFLE-UPL.txt

import glob
import os
import shutil
from common import *

args = os.sys.argv
args.pop(0)

if len(args) == 0:
    raise RuntimeError("Missing INPUT_DIR")
input_dir = args.pop(0)

if len(args) == 0:
    raise RuntimeError("Missing OUTPUT_DIR")
output_dir = args.pop(0)

if len(args) == 0:
    raise RuntimeError("Missing TEST_NAME")
test_name = args.pop(0)

if len(args) > 0:
    extra_data = args.pop(0)
else:
    extra_data = None

if len(args) != 0:
    raise RuntimeError("Unexpected argument")

if os.path.exists(output_dir) and (not os.path.isdir(output_dir)):
    raise RuntimeError("Output path exists but is not a directory")

# clean output directory
if os.path.isdir(output_dir):
    for file in os.listdir(output_dir):
        path = os.path.join(output_dir, file)
        if os.path.isdir(path):
            shutil.rmtree(path) # only remove directories
        else:
            os.remove(path) # only remove files
else:
    os.makedirs(output_dir, exist_ok=True)

if extra_data == "json":
    for file in [os.path.basename(p) for p in glob.glob(os.path.join(input_dir,"*.json"))]:
        shutil.copyfile(
                os.path.join(input_dir, file),
                os.path.join(output_dir, "{}.expected".format(file)))
elif extra_data == "python" or extra_data == "java":
    shutil.copyfile(
            os.path.join(input_dir, "{}-{}.out".format(test_name, extra_data)),
            os.path.join(output_dir, "{}-{}.out.expected".format(test_name, extra_data)))
elif extra_data == "provenance":
    shutil.copyfile(
            os.path.join(input_dir, "{}.in".format(test_name)),
            os.path.join(output_dir, "{}.in".format(test_name)))

shutil.copyfile(
        os.path.join(input_dir, "{}.out".format(test_name)),
        os.path.join(output_dir, "{}.out.expected".format(test_name)))

shutil.copyfile(
        os.path.join(input_dir, "{}.err".format(test_name)),
        os.path.join(output_dir, "{}.err.expected".format(test_name)))

csvs = [os.path.basename(p) for p in glob.glob(os.path.join(input_dir,"*.csv"))]

with open(os.path.join(output_dir, "num.expected"), "w") as num_file:
    num_file.write("{}\n".format(len(csvs)))

for file in csvs:
    shutil.copyfile(
            os.path.join(input_dir, file),
            os.path.join(output_dir, "{}.expected".format(file)))

owd = os.getcwd()
try:
    os.chdir(output_dir)
    for file in glob.glob("*.expected"):
        sort_file(file)
finally:
    os.chdir(owd)




# Souffle - A Datalog Compiler
# Copyright (c) 2022 The Souffle Developers. All rights reserved
# Licensed under the Universal Permissive License v 1.0 as shown at:
# - https://opensource.org/licenses/UPL
# - <souffle root>/licenses/SOUFFLE-UPL.txt

import os
import argparse
import pathlib
import subprocess

parser = argparse.ArgumentParser(description="Redirect standard streams")
parser.add_argument('--in', dest='in_file')
parser.add_argument('--out', dest='out_file')
parser.add_argument('--err', dest='err_file')
parser.add_argument('command', type=lambda p: pathlib.Path(p).absolute())
parser.add_argument('arguments', nargs=argparse.REMAINDER)

args = parser.parse_args()

if args.in_file:
    stdin = open(args.in_file)
else:
    stdin = None

if args.out_file:
    stdout = open(args.out_file, "w")
else:
    stdout = None

if args.err_file:
    stderr = open(args.err_file, "w")
else:
    stderr = None

status = subprocess.run([args.command] + args.arguments, stdin=stdin, stdout=stdout, stderr=stderr)

if stdout:
    stdout.close()

if stderr:
    stderr.close()

if status.returncode != 0 and args.err_file:
    with open(args.err_file, "r") as f:
        os.sys.stderr.write(f.read())

os.sys.exit(status.returncode)



# Souffle - A Datalog Compiler
# Copyright (c) 2022 The Souffle Developers. All rights reserved
# Licensed under the Universal Permissive License v 1.0 as shown at:
# - https://opensource.org/licenses/UPL
# - <souffle root>/licenses/SOUFFLE-UPL.txt

import os
import glob
import subprocess
from common import *

args = os.sys.argv
args.pop(0)

if len(args) == 0:
    raise RuntimeError("Missing INPUT_DIR")

input_dir = args.pop(0)
extra_data = None
binary = None

if len(args) > 0:
    arg = args.pop(0)
    if arg == "":
        extra_data = None
    elif arg == "sqlite3" or arg == "gzip":
        extra_data = arg
        if len(args) == 0:
            raise RuntimeError("Missing BINARY")
        binary = args.pop(0)
    elif arg == "json":
        extra_data = arg
        if len(args) > 0:
            args.pop(0) # unused last argument
    else:
        raise RuntimeError("Unknown processing type {}".format(arg))

if len(args) != 0:
    raise RuntimeError("Unexpected argument")

if extra_data == "gzip":
    extra_file_pattern = "*.gz.output"
elif extra_data == "sqlite3":
    extra_file_pattern = "*.sqlite.output"
elif extra_data == "json":
    extra_file_pattern = "*.json"
else:
    extra_file_pattern = None

if extra_file_pattern:
    extra_files = glob.glob(extra_file_pattern)

    for file in extra_files:
        if extra_data == "gzip":
            generated_file = os.path.basename(file).rstrip(".gz.output")
            with open(generated_file, "w") as generated:
                subprocess.run([binary, "-d", "-c", file], check=True, stdout=generated)
        elif extra_data == "sqlite3":
            generated_file = "{}.csv".format(os.path.basename(file).rstrip(".sqlite.output"))
            script = os.path.join(input_dir, "{}.script".format(file))
            with open(generated_file, "w") as generated:
                subprocess.run([binary,"-batch","-init",script,file,""], check=True, stdout=generated)
        elif extra_data == "json":
            generated_file = file

        sort_file(generated_file)
        compare_sorted_file(generated_file)

generated_csvs = glob.glob("*.csv")

with open("num.generated", "w") as num_file:
    num_file.write("{}\n".format(len(generated_csvs)))

compare_files("num.expected", "num.generated")

csvs = list(map(
        lambda file: os.path.basename(file).rstrip(".expected.sorted"),
        glob.glob("*.csv.expected.sorted")))

for csv_file in csvs:
    sort_file(csv_file)
    compare_sorted_file(csv_file)




"""
Souffle - A Datalog Compiler
Copyright (c) 2019, The Souffle Developers. All rights reserved
Licensed under the Universal Permissive License v 1.0 as shown at:
- https://opensource.org/licenses/UPL
- <souffle root>/licenses/SOUFFLE-UPL.txt
"""

import SwigInterface
import sys
p = SwigInterface.newInstance('insert_for')
p.loadAll(sys.argv[1])
p.run()
p.printAll('.')
p.thisown = 1
del p
